%% Demo: Robust PCA for background-foreground separation of videos
% Task:
%   We apply Alternating Direction Method of Multipliers (ADMM) to solve this model,
% where the augmented Lagrangian is given by:
%   L(A,B,M,Y) = |A|_nuc + lambda*|B|_l1 + delta{|M-Z|_fro <= eps} +
%   <Y, A+B-M> + 0.5*rho*|A+B-M|_fro^2
%   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   


%% problem setup
clear all
close all hidden

rng(1)

addpath('./util');
    
% number of frames (max 200), use small number for testing.
nFrame=20;

data_formation


%% ADMM parameters
lambda = .5/sqrt(n1); % parameter of l1 norm on B
eps = .01;            % threshold for ||M-Z||_fro
rho = 1;              % parameter in ADMM
max_it = 100;         % max number of iterations

% array to record the change of energy.
ener = zeros(1,max_it);


%% initialization
A = Z;
B = zeros(n1,n2);
M = Z;
Y = zeros(n1,n2);
% rank estimate of A
rankA_est = min(n1,n2);


%% ADMM solver
for iter = 1:max_it
    % update A:
    prox_A = -B+M-Y/rho;
    [U,S,V] = svds(prox_A, rankA_est);
    S = max(0,diag(S)-1/rho);
    A = U*diag(S)*V';
    rankA_est = min(sum(S>0)+5, min(n1,n2));
    
    % update B:
    prox_B = M-A-Y/rho;
    B = zeros(size(B));
    B(prox_B<-lambda/rho) = prox_B(prox_B<-lambda/rho) + lambda/rho;
    B(prox_B>lambda/rho) = prox_B(prox_B>lambda/rho) - lambda/rho;
    
    % update M:
    prox_M = A+B+Y/rho;
    M = prox_M-Z;
    M = Z + min(1, eps/max(eps*1e-3,norm(M(:))))*M;
    
    % update Y:
    Y = Y+rho*(A+B-M);
    
    % stats
    ener(iter) = sum(sum(S)) + lambda*sum(sum(abs(B)))...
        + dot(Y(:),A(:)+B(:)-M(:)) + .5*rho*sum((A(:)+B(:)-M(:)).^2);
    if mod(iter,5)==0
        fprintf('it %d: augmented Lagrangian= %.8g, rankA_est= %d\n', ...
            iter, ener(iter), rankA_est)
    end
    
end


%% show result
A = imnorm(reshape(A,[z1 z2 z3]));
B = imnorm(reshape(B,[z1 z2 z3]));
Z = imnorm(reshape(Z,[z1 z2 z3]));
figure(1)
for i = 1:nFrame
   subplot(1,3,1)
   subimage(A(:,:,i));
   title('background')
   subplot(1,3,2)
   subimage(B(:,:,i));
   title('foreground')
   subplot(1,3,3)
   subimage(Z(:,:,i));
   title('original');
   pause(1/35)
end
