# Demo: Robust PCA for background-foreground separation of videos
#
# Original video content taken from:
#   http://perception.i2r.a-star.edu.sg/bk_model/bk_index.html
#
# References:
#   - Candes & Li & Ma & Wring, Robust principal component analysis?,
#   J. ACM, 2011.
#   - Tao & Yuan, Recovering Low-Rank and Sparse Components of Matrices from
#   Incomplete and Noisy Observations, SIOPT 2011.
#
# Description:
#   The task is to extract the background and the foreground of video. Assume
# raw video frames are stacked in a matrix $Z \in \mathbb{R}^{M*N}$, where $N$
# is the number of frames in the video and $M$ is the number of pixels in each
# frame. We build a background matrix $A$ and a foreground matrix $B$ of the
# same size as $Z$, and expect matrix $A$ to be low-rank and matrix $B$ to be
# sparse. Meanwhile, we allow the additive decomposition $Z=A+B$ to be inexact
# up to small noise. To this end, the variational model is formulated as
# follows:
#
# min_{A,B,M} |A|_nuc + lambda*|B|_l1 + delta{|M-Z|_fro<=eps} + delta{A+B-M=0},
#
# where |A|_nuc is the nuclear norm of matrix A (i.e. the sum of all singular
# values of A), |B|_l1 is the l1-norm of (vectorized) B, delta{ } is the
# indicator function, | |_fro is the Frobenius norm.
#
# Task:
#   We shall apply Alternating Direction Method of Multipliers (ADMM) to solve
# this model, where the augmented Lagrangian is given by:
#   L(A,B,M,Y) = |A|_nuc + lambda*|B|_l1 + delta{|M-Z|_fro <= eps} +
#   <Y, A+B-M> + 0.5*rho*|A+B-M|_fro^2
#
# #############################################################################


import os
import numpy as np
# import scipy.sparse.linalg as splinalg
import matplotlib.pyplot as plt
from PIL import Image


def im2arr(img):
    """pillow grayscale image to normalized numpy array"""
    return np.array(img, dtype=np.float64) / 255.0


def imnorm(imgmat):
    """normalize image from matrix representation"""
    imgmat = (imgmat - np.min(imgmat)) * 255. / \
        (np.max(imgmat) - np.min(imgmat) + 10e-20)
    return Image.fromarray(imgmat.astype(np.uint8)).convert('L')


def mat2vec(m):
    """flatten matrix m"""
    return m.T.flatten()


def vec2mat(v, shape):
    h, w = shape
    return v.reshape(w, h).T


def get_synthetic_data(nFrame):
    # parameters
    h, w = 20, 25
    n1, n2 = h * w, nFrame       # matrix size
    r = min(n1, n2) // 10   # presumed matrix rank
    s = (n1 * n2) // 10     # presumed matrix sparsity
    noise = 0.01            # noise level
    # true low-rank matrix A_0
    A_0 = np.random.randn(n1, r) @ np.random.randn(r, n2)
    A_max = np.sqrt(max(n1, n2))
    # true sparse matrix B_0
    B_0 = ((np.random.rand(n1, n2) < 0.5) - 0.5) * 2 * A_max
    B_0[np.unravel_index(np.random.permutation(B_0.size)[:n1*n2-s],
                         B_0.shape)] = 0
    # build up the noise output
    Z = A_0 + B_0
    Z += noise * np.random.randn(*Z.shape)
    return Z, h, w


# Original video content taken from:
# http://perception.i2r.a-star.edu.sg/bk_model/bk_index.html
def get_airport_data(nFrame):
    # parameters
    data_dir = 'video_airport/'
    initFrame = 1000
    w, h = Image.open(os.path.join(data_dir, 'airport1000.bmp')).size
    # read data
    img_arrays = []
    for i in range(initFrame, initFrame + nFrame):
        img_name = os.path.join(data_dir, 'airport{}.bmp'.format(i))
        img = mat2vec(im2arr(Image.open(img_name).convert('L')))
        img_arrays.append(img)
    return np.stack(img_arrays, axis=1), h, w


def plot_augmented_lagrangians(energies):
    plt.figure('augmented lagrangian')
    plt.plot(range(len(energies)), energies)
    plt.show()


def display_input(Z, m, n):
    """display input video frame-by-frame"""
    plt.figure('Input video')
    n1, n2 = Z.shape
    for i in range(n2):
        plt.imshow(imnorm(vec2mat(Z[:, i], (m, n))))
        plt.title('original ({})'.format(i))
        plt.draw()
        plt.pause(1.0 / 12)  # show at 12 fps
        plt.clf()
    plt.close()


def display_output(A, B, Z, m, n):
    """display output frame-by-frame showing the decomposition"""
    plt.figure('Foreground/background separation')
    n1, n2 = Z.shape
    for i in range(n2):
        plt.imshow(imnorm(np.concatenate(
                [vec2mat(A[:, i], (m, n)), vec2mat(B[:, i], (m, n)),
                 vec2mat(Z[:, i], (m, n))], axis=1)))
        plt.title('background / foreground / original ({})'.format(i))
        plt.draw()
        plt.pause(1.0 / 12)  # show at 12 fps
        plt.clf()


def update_A(B, M, Y, rho, rankA_est):
    prox_A = - B + M - Y/rho
    # U, S, Vt = splinalg.svds(prox_A, rankA_est)  # actually slower ...
    U, S, Vt = np.linalg.svd(prox_A, full_matrices=False)
    S = np.maximum(0, S - 1/rho)
    A = U @ np.diag(S) @ Vt
    rankA_est = min(np.sum(S > 0) + 5, min(n1, n2) - 1)
    return A, rankA_est, S


def update_B(A, M, Y, rho, lamb):
    prox_B = M - A - Y/rho
    B_new = np.zeros_like(B)
    B_new[prox_B < -lamb/rho] = prox_B[prox_B < -lamb/rho] + lamb/rho
    B_new[prox_B > lamb/rho] = prox_B[prox_B > lamb/rho] - lamb/rho
    return B_new


def update_M(A, B, Y, Z, rho, eps):
    prox_M = A + B + Y/rho
    M_new = prox_M - Z
    M_new = Z + min(1, eps/max(eps*1e-3, np.linalg.norm(M_new))) * M_new
    return M_new


def update_Y(A, B, M, Y, rho):
    return Y + rho * (A + B - M)


def get_augmented_lagrangian(A, B, M, Y, rho, S):
    return np.sum(S) + lamb*np.sum(np.abs(B)) \
        + np.sum(Y * (A + B - M)) + 0.5*rho*np.sum((A+B-M) ** 2)
    # return np.sum(np.maximum(0, S-1/rho)) + lamb*np.sum(np.abs(B)) \
    #     + np.sum(Y * (A + B - M)) + 0.5*rho*np.sum((A+B-M) ** 2)


if __name__ == '__main__':
    # get input data
    use_synthetic_data = False
    nFrame = 100  # should be <= 200 for the real airport data
    Z, m, n = (get_airport_data, get_synthetic_data
               )[use_synthetic_data](nFrame)
    n1, n2 = Z.shape

    # display input data
    display_input(Z, m, n)

    # algorithm parameters
    lamb = 0.5 / np.sqrt(n1)
    eps = 0.01
    rho = 1.0
    max_it = 200

    # initialization
    A = None
    rankA_est = min(n1, n2) - 1  # initial rank estimate of A
    B = np.zeros([n1, n2])
    M = Z
    Y = np.zeros([n1, n2])
    S = None  # store eigenvalues of A for augmented Lagrangian
    energies = []  # stores augmented lagrangian

    # ADMM solver
    for it in range(max_it):
        # update A:
        A, rankA_est, S = update_A(B, M, Y, rho, rankA_est)
        # update B:
        B = update_B(A, M, Y, rho, lamb)
        # update M:
        M = update_M(A, B, Y, Z, rho, eps)
        # update Y:
        Y = update_Y(A, B, M, Y, rho)
        # update energies
        energy = get_augmented_lagrangian(A, B, M, Y, rho, S)
        energies.append(energy)
        # output status
        print("it {}: augmented Lagrangian= {:.8f}, rankA_est= {}".format(
                it+1, energy, rankA_est))

    # visualize results
    plot_augmented_lagrangians(energies)
    display_output(A, B, Z, m, n)
